import uuid
from numpy import float32, float64, int16, int32, int64
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import (
    MetricsSet,
    MetricType
)
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)
from typing import Dict
from .datadog_config import DatadogConfig

try:
    # Datadog uses aiosonic
    from aiosonic import HTTPClient, TCPConnector, Timeouts
    from datadog_api_client import (
        AsyncApiClient, 
        Configuration
    )
    from datadog_api_client.v2.api.metrics_api import (MetricsApi, MetricPayload)
    from datadog_api_client.v2.model.metric_series import MetricSeries
    from datadog_api_client.v2.model.metric_point import MetricPoint

    from datadog_api_client.v1.api.events_api import EventsApi
    from datadog_api_client.v1.api.events_api import EventCreateRequest
    has_connector = True

except Exception:
    datadog = None
    has_connector = False

from datetime import datetime
from typing import List


class Datadog:

    def __init__(self, config: DatadogConfig) -> None:
        self.datadog_api_key = config.api_key
        self.datadog_app_key = config.app_key
        self.event_alert_type = config.event_alert_type or 'info'
        self.device_name = config.device_name or 'hedra'
        self.priority = config.priority
        self.custom_fields = config.custom_fields or {}

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self.types_map = {
            'total': 'count',
            'succeeded': 'count',
            'failed': 'count',
            'actions_per_second': 'gauge',
            'median': 'gauge',
            'mean': 'gauge',
            'variance': 'gauge',
            'stdev': 'gauge',
            'minimum': 'gauge',
            'maximum': 'gauge',
            **self.custom_fields
        }
        
        self._datadog_api_map = {
            'count': 1,
            'rate': 2,
            'gauge': 3,
            'histogram': 4,
        }

        self._config = None
        self._client = None
        self.events_api = None
        self.metrics_api = None
        self.metric_types_map = {
            MetricType.COUNT: 'count',
            MetricType.RATE: 'rate',
            MetricType.DISTRIBUTION: 'histogram',
            MetricType.SAMPLE: 'gauge'
        }

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to Datadogg API')

        self._config = Configuration()
        self._config.api_key["apiKeyAuth"] = self.datadog_api_key
        self._config.api_key["appKeyAuth"] = self.datadog_app_key

        self._client = AsyncApiClient(self._config)

        # Datadog's implementation of aiosonic's HTTPClient lacks a lot
        # of configurability, incuding actually being able to set request timeouts
        # so we substitute our own implementation.

        tcp_connection = TCPConnector(timeouts=Timeouts(sock_connect=30))
        self._client.rest_client._client = HTTPClient(tcp_connection)

        self.events_api = EventsApi(self._client)
        self.metrics_api = MetricsApi(self._client)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to Datadogg API')

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to Datadog API')

        metrics_sets: List[SessionMetricsCollection] = []
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.system_metrics_set_id}')
            
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                metrics_sets.append(monitor_metrics)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                metrics_sets.append(monitor_metrics)

        system_session_metrics_series: List[MetricSeries] = []

        for metrics_set in metrics_sets:

            tags = [
                f'group:{metrics_set.group}'
            ]

            for field, value in metrics_set.stats.items():

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Session System Metric - {metrics_set.name}:{metrics_set.group}:{field}')

                metric_type = metrics_set.types_map.get(field)
                datadog_metric_type = self._datadog_api_map.get(metric_type)

                if isinstance(value, (int, int16, int32, int64)):
                    value = int(value)

                elif isinstance(value, (float, float32, float64)):
                    value = float(value)

                series = MetricSeries(
                    f'{metrics_set.name}_{metrics_set.group}_{field}', 
                    [MetricPoint(
                        timestamp=int(datetime.now().timestamp()),
                        value=value
                    )],
                    type=datadog_metric_type,
                    tags=tags
                )

                system_session_metrics_series.append(series)
                
        await self.metrics_api.submit_metrics(MetricPayload(system_session_metrics_series))
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Session System Metrics to Datadog API')

    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to Datadog API')

        metrics_sets: List[SystemMetricsCollection] = []
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.system_metrics_set_id}')
            
            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    metrics_sets.append(stage_mb_per_vu_metrics)

        system_stage_metrics_series: List[MetricSeries] = []

        for metrics_set in metrics_sets:

            tags = [
                f'group:{metrics_set.group}'
            ]

            for field, value in metrics_set.stats.items():

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Stage System Metric - {metrics_set.name}:{metrics_set.group}:{field}')

                metric_type = metrics_set.types_map.get(field)
                datadog_metric_type = self._datadog_api_map.get(metric_type)

                if isinstance(value, (int, int16, int32, int64)):
                    value = int(value)

                elif isinstance(value, (float, float32, float64)):
                    value = float(value)

                series = MetricSeries(
                    f'{metrics_set.name}_{metrics_set.group}_{field}', 
                    [MetricPoint(
                        timestamp=int(datetime.now().timestamp()),
                        value=value
                    )],
                    type=datadog_metric_type,
                    tags=tags
                )

                system_stage_metrics_series.append(series)
                
        await self.metrics_api.submit_metrics(MetricPayload(system_stage_metrics_series))
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Stage System Metrics to Datadog API')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to Datadog API')

        streams_series = []
        for stage_name, stream in stream_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stream - {stage_name}:{stream.stream_set_id}')
            
            for group_name, group in stream.grouped.items():
                tags = [
                    f'stage:{stage_name}',
                    f'group:{group}'
                ]

                for field, value in group.items():

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Stream Metric - {stage_name}:{group_name}:{field}')

                    metric_type = stream.types_map.get(field)
                    datadog_metric_type = self._datadog_api_map.get(metric_type)

                    if isinstance(value, (int, int16, int32, int64)):
                        value = int(value)

                    elif isinstance(value, (float, float32, float64)):
                        value = float(value)

                    series = MetricSeries(
                        f'{stage_name}_{group}_{field}', 
                        [MetricPoint(
                            timestamp=int(datetime.now().timestamp()),
                            value=value
                        )],
                        type=datadog_metric_type,
                        tags=tags
                    )

                    streams_series.append(series)
                
        await self.metrics_api.submit_metrics(MetricPayload(streams_series))
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to Datadog API')        

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to Datadog API')

        experiments_series = []
        for experiment in experiment_metrics.experiment_summaries:

            experiment_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Experiment - {experiment.experiment_name}:{experiment_id}')

            tags = [
                f'experiment_name:{experiment.experiment_name}',
                f'experiment_randomized:{experiment.experiment_randomized}'
            ]

            for field, value in experiment.stats:

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Experiments - {experiment.experiment_name}:{field}')

                metric_type = experiment.types_map.get(field)
                datadog_metric_type = self._datadog_api_map.get(metric_type.value)

                series = MetricSeries(
                    f'{experiment.experiment_name}_{field}', 
                    [MetricPoint(
                        timestamp=int(datetime.now().timestamp()),
                        value=value
                    )],
                    type=datadog_metric_type,
                    tags=tags
                )

                experiments_series.append(series)
                
        await self.metrics_api.submit_metrics(MetricPayload(experiments_series))
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to Datadog API')
    
    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to Datadog API')

        variant_series = []
        for variant in experiment_metrics.variant_summaries:

            variant_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Variants - {variant.variant_name}:{variant_id}')

            tags = [
                f'variant_name:{variant.variant_name}',
                f'variant_experiment:{variant.variant_experiment}',
                f'variant_distribution:{variant.variant_distribution}',
            ]

            for field, value in variant.stats:

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Variant - {variant.variant_name}:{field}')

                metric_type = variant.types_map.get(field)
                datadog_metric_type = self._datadog_api_map.get(metric_type.value)

                series = MetricSeries(
                    f'{variant.variant_name}_{field}', 
                    [MetricPoint(
                        timestamp=int(datetime.now().timestamp()),
                        value=value
                    )],
                    type=datadog_metric_type,
                    tags=tags
                )

                variant_series.append(series)
                
        await self.metrics_api.submit_metrics(MetricPayload(variant_series))
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to Datadog API')
    
    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to Datadog API')

        mutation_series = []
        for mutation in experiment_metrics.mutation_summaries:

            mutation_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Mutations - {mutation.mutation_name}:{mutation_id}')

            tags = [
                f'mutation_name:{mutation.mutation_name}',
                f'mutation_experiment_name:{mutation.mutation_experiment_name}',
                f'mutation_variant_name:{mutation.mutation_variant_name}',
                f'mutation_targets:{mutation.mutation_targets}',
                f'mutation_type:{mutation.mutation_type}',
            ]

            for field, value in mutation.stats:

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Mutation - {mutation.mutation_name}:{field}')

                metric_type = mutation.types_map.get(field)
                datadog_metric_type = self._datadog_api_map.get(metric_type.value)

                series = MetricSeries(
                    f'{mutation.mutation_name}_{field}', 
                    [MetricPoint(
                        timestamp=int(datetime.now().timestamp()),
                        value=value
                    )],
                    type=datadog_metric_type,
                    tags=tags
                )

                mutation_series.append(series)
                
        await self.metrics_api.submit_metrics(MetricPayload(mutation_series))
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to Datadog API')

    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Datadog API')

        for event in events:

            tags = {
                f'{tag.name}:{tag.value}' for tag in event.tags
            }

            await self.events_api.create_event(
                EventCreateRequest(
                    title=event.name,
                    text=event.serialize(),
                    alert_type=self.event_alert_type,
                    aggregation_key=event.type,
                    device_name=self.device_name,
                    date_happened=datetime.now().strftime('%Y-%m-%dT%H:%M:%S.Z'),
                    priority=self.priority,
                    tags=tags,
                    host=event.hostname
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Datadog API')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Datadog API')

        metrics_series = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            tags = [
                f'{tag.name}:{tag.value}' for tag in metrics_set.tags
            ]

            for field, value in metrics_set.common_stats.items():

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Shared Metric - {metrics_set.name}:common:{field}')

                metric_type = self.types_map.get(field)
                datadog_metric_type = self._datadog_api_map.get(metric_type)

                if isinstance(value, (int, int16, int32, int64)):
                    value = int(value)

                elif isinstance(value, (float, float32, float64)):
                    value = float(value)

                series = MetricSeries(
                    f'{metrics_set.name}_{field}', 
                    [MetricPoint(
                        timestamp=int(datetime.now().timestamp()),
                        value=value
                    )],
                    type=datadog_metric_type,
                    tags=[
                        *tags,
                        f'metric_stage:{metrics_set.stage}',
                        f'group:common'
                    ]
                )

                metrics_series.append(series)
                
        await self.metrics_api.submit_metrics(MetricPayload(metrics_series))
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Datadog API')        

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Datadog API')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            tags = [
                f'{tag.name}:{tag.value}' for tag in metrics_set.tags
            ]

            metrics_series = []
            for group_name, group in metrics_set.groups.items():

                for field, value in group.stats.items():

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Metric - {metrics_set.name}:{group_name}:{field}')

                    metric_type = self.types_map.get(field)
                    datadog_metric_type = self._datadog_api_map.get(metric_type)

                    if isinstance(value, (int, int16, int32, int64)):
                        value = int(value)

                    elif isinstance(value, (float, float32, float64)):
                        value = float(value)

                    series = MetricSeries(
                        f'{metrics_set.name}_{field}', 
                        [MetricPoint(
                            timestamp=int(datetime.now().timestamp()),
                            value=float(value)
                        )],
                        type=datadog_metric_type,
                        tags=[
                            *tags,
                            f'metric_stage:{metrics_set.stage}',
                            f'group:{group_name}'
                        ]
                    )

                    metrics_series.append(series)

                for quantile_name, quantile_value in group.quantiles.items():

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Metric Quantile - {metrics_set.name}:{group_name}:{quantile_name}th Quantile')

                    if isinstance(quantile_value, (int, int16, int32, int64)):
                        quantile_value = int(quantile_value)

                    elif isinstance(quantile_value, (float, float32, float64)):
                        quantile_value = float(quantile_value)

                    datadog_metric_type = self._datadog_api_map.get('gauge')
                    series = MetricSeries(
                        f'{metrics_set.name}_{quantile_name}', 
                        [MetricPoint(
                            timestamp=int(datetime.now().timestamp()),
                            value=float(quantile_value)
                        )],
                        type=datadog_metric_type,
                        tags=[
                            *tags,
                            f'metric_stage:{metrics_set.stage}',
                            f'group:{group_name}'
                        ]
                    )

                    metrics_series.append(series)
     
            await self.metrics_api.submit_metrics(MetricPayload(metrics_series))

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Datadog API')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to Datadog API')

        metrics_series = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            tags = [
                f'{tag.name}:{tag.value}' for tag in metrics_set.tags
            ]

            for custom_metric_name, custom_metric in metrics_set.custom_metrics.items():

                metric_type = self.metric_types_map.get(
                    custom_metric.metric_type,
                    'gauge'
                )

                datadog_metric_type = self._datadog_api_map.get(metric_type)

                series = MetricSeries(
                    f'{metrics_set.name}_{custom_metric_name}',
                    [MetricPoint(
                        timestamp=int(datetime.now().timestamp()),
                        value=custom_metric.metric_name
                    )],
                    type=datadog_metric_type,
                    tags=[
                        *tags,
                        f'metric_stage:{metrics_set.stage}',
                        'group:custom'
                    ]
                )

                metrics_series.append(series)

        await self.metrics_api.submit_metrics(MetricPayload(metrics_series))

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Datadog API')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Datadog API')

        error_series = [] 
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            tags = [
                f'{tag.name}:{tag.value}' for tag in metrics_set.tags
            ]

            for error in metrics_set.errors:
                
            
                error_message = error.get('error_message')

                series = MetricSeries(
                    f'{metrics_set.name}_errors',
                    [MetricPoint(
                        timestamp=int(datetime.now().timestamp()),
                        value=int(error.get('count'))
                    )],
                    type=self._datadog_api_map.get('count'),
                    tags=[
                        *tags,
                        f'metric_stage:{metrics_set.stage}',
                        f'error_message:{error_message}'
                    ]
                )
               
                error_series.append(series)

        await self.metrics_api.submit_metrics(MetricPayload(error_series))

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Datadog API')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')


from typing import Any, Dict, Optional
from hedra.data.connectors.common.connector_type import ConnectorType
from pydantic import (
    BaseModel, 
    StrictStr, 
    StrictInt,
    StrictBool
)


class KafkaConnectorConfig(BaseModel):
    host: StrictStr='localhost:9092'
    client_id: StrictStr='hedra'
    topic: StrictStr
    partition: StrictInt=0
    compression_type: Optional[StrictStr]
    timeout: StrictInt=1000
    idempotent: StrictBool=True
    options: Dict[StrictInt, Any]={}
    reporter_type: ConnectorType=ConnectorType.Kafka


import functools
from hedra.distributed.models.http import (
    Request,
    Limit
)
from pydantic import BaseModel
from typing import (
    Optional, 
    List, 
    Literal,
    Dict,
    Callable,
    Tuple,
    TypeVar,
    Any
)

T = TypeVar('T')


def endpoint(
    path: Optional[str]="/",
    methods: List[
        Literal[
            "GET",
            "HEAD",
            "OPTIONS",
            "POST",
            "PUT",
            "PATCH",
            "DELETE",
            "TRACE"
        ]
    ]=["GET"],
    responses: Optional[
        Dict[
            int,
            BaseModel
        ]
    ]=None,
    serializers: Optional[
        Dict[
            int,
            Callable[
                ...,
                str
            ]
        ]
    ]=None,
    middleware: Optional[
        List[
            Callable[
                [
                    Request
                ],
                Tuple[
                    Any,
                    int,
                    bool
                ]
            ]
        ]
    ]=None,
    response_headers: Optional[Dict[str, str]]=None,
    limit: Optional[Limit]=None
):

    def wraps(func):

        func.server_only = True
        func.path = path
        func.methods = methods
        func.as_http = True
        
        func.response_headers = response_headers or {}
        func.responses = responses
        func.serializers = serializers
        func.limit = limit
        
        if middleware:
            @functools.wraps(func)
            async def middleware_decorator(
                *args,
                **kwargs
            ):
            
                run_next = True

                _, request = args

                for middleware_func in middleware:
                    response, run_next = await middleware_func(request)

                    if run_next is False:
                        return response
                    

                return await func(*args, **kwargs)
            
            return middleware_decorator
        

        else:
            @functools.wraps(func)
            def decorator(
                *args,
                **kwargs
            ):
                return func(*args, **kwargs)
            
            return decorator
    
    return wraps



import re
import uuid
from typing import List, Dict
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import (
    MetricsSet,
    MetricType
)
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)
from .statsd_config import StatsDConfig


try:
    from aio_statsd import StatsdClient
    has_connector = True

except Exception:
    StatsdClient = None
    has_connector = False


class StatsD:

    def __init__(self, config: StatsDConfig) -> None:
        self.host = config.host
        self.port = config.port

        self.connection = StatsdClient(
            host=self.host,
            port=self.port
        )

        self.types_map = {
            'total': 'count',
            'succeeded': 'count',
            'failed': 'count',
            'actions_per_second': 'gauge',
            'median': 'gauge',
            'mean': 'gauge',
            'variance': 'gauge',
            'stdev': 'gauge',
            'minimum': 'gauge',
            'maximum': 'gauge',
            'quantiles': 'gauge'
        }

        self._update_map = {
            'count': self.connection.counter,
            'gauge': self.connection.gauge,
            'increment': self.connection.increment,
            'sets': self.connection.sets,
            'histogram': lambda: NotImplementedError('StatsD does not support histograms.'),
            'distribution': lambda: NotImplementedError('StatsD does not support distributions.'),
            'timer': self.connection.timer

        }

        self.stat_type_map = {
            MetricType.COUNT: 'count',
            MetricType.DISTRIBUTION: 'gauge',
            MetricType.RATE: 'gauge',
            MetricType.SAMPLE: 'gauge'
        }

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self.statsd_type = 'StatsD'

    async def connect(self):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to {self.statsd_type} at - {self.host}:{self.port}')
        await self.connection.connect()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to {self.statsd_type} at - {self.host}:{self.port}')

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to {self.statsd_type}')

        metrics_sets: List[SessionMetricsCollection] = []
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Preparing Session System Metrics Set - {metrics_set.system_metrics_set_id}')
  
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                metrics_sets.append(monitor_metrics)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                metrics_sets.append(monitor_metrics)

        for metrics_set in metrics_sets:

            for metric_field, metric_value in metrics_set.record.items():
                
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metric Set - {metrics_set.name}:{metrics_set.group}:{metric_field}')
                
                update_type = metrics_set.types_map.get(metric_field)
                stat_type = self.stat_type_map.get(update_type)
                update_function = self._update_map.get(stat_type)
                
                update_function(
                    f'{metrics_set.name}_{metrics_set.group}_{metric_field}',
                    metric_value
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Session System Metrics to {self.statsd_type}')

    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to {self.statsd_type}')

        metrics_sets: List[SystemMetricsCollection] = []
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Preparing Stage System Metrics Set - {metrics_set.system_metrics_set_id}')
  
            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    metrics_sets.append(stage_mb_per_vu_metrics)

        for metrics_set in metrics_sets:

            for metric_field, metric_value in metrics_set.record.items():
                
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metric Set - {metrics_set.name}:{metrics_set.group}:{metric_field}')
                
                update_type = metrics_set.types_map.get(metric_field)
                stat_type = self.stat_type_map.get(update_type)
                update_function = self._update_map.get(stat_type)
                
                update_function(
                    f'{metrics_set.name}_{metrics_set.group}_{metric_field}',
                    metric_value
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Stage System Metrics to {self.statsd_type}')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to {self.statsd_type}')

        for stage_name, stream in stream_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stream - {stage_name}:{stream.stream_set_id}')

            for group_name, group in stream.grouped.items():

                for metric_field, metric_value in group.items():
                    
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stream Metric - {stage_name}:{group_name}:{metric_field}')
                    
                    update_type = stream.types_map.get(group_name)
                    stat_type = self.stat_type_map.get(update_type)
                    update_function = self._update_map.get(stat_type)
                    
                    update_function(
                        f'{stage_name}_stream_{group_name}_{metric_field}',
                        metric_value
                    )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Streams to {self.statsd_type}')

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to {self.statsd_type}')

        for experiment in experiment_metrics.experiment_summaries:

            experiment_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Experiment - {experiment.experiment_name}:{experiment_id}')
            
            for field, value in experiment.stats:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Experiment field - {experiment.experiment_name}:{field}')

                update_type = experiment.types_map.get(field)
                update_function = self._update_map.get(update_type.value)

                update_function(
                    f'{experiment.experiment_name}_{field}', value
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to {self.statsd_type}')

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to {self.statsd_type}')

        for variant in experiment_metrics.variant_summaries:

            variant_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Variant - {variant.variant_name}:{variant_id}')
            
            for field, value in variant.stats:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Variants field - {variant.variant_name}:{field}')

                update_type = variant.types_map.get(field)
                update_function = self._update_map.get(update_type.value)

                update_function(
                    f'{variant.variant_name}_{field}', value
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to {self.statsd_type}')

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to {self.statsd_type}')

        for mutation in experiment_metrics.mutation_summaries:

            mutation_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Mutation - {mutation.mutation_name}:{mutation_id}')
            
            for field, value in mutation.stats:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Mutatio field - {mutation.mutation_name}:{field}')

                update_type = mutation.types_map.get(field)
                update_function = self._update_map.get(update_type.value)

                update_function(
                    f'{mutation.mutation_name}_{field}', value
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to {self.statsd_type}')

    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to {self.statsd_type}')

        for event in events:
            time_update_function = self._update_map.get('gauge')
            time_update_function(f'{event.name}_time', event.time)
            
            if event.success:
                success_update_function = self._update_map.get('count')
                success_update_function(f'{event.name}_success', 1)
            
            else:
                failed_update_function = self._update_map.get('count')
                failed_update_function(f'{event.name}_failed', 1)
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to {self.statsd_type}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to {self.statsd_type}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            for field, value in metrics_set.common_stats.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metric - {metrics_set.name}:common:{field}')

                update_type = self.types_map.get(field)
                update_function = self._update_map.get(update_type)

                update_function(
                    f'{metrics_set.name}_common_{field}', value
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to {self.statsd_type}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to {self.statsd_type}')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():

                metric_record = {**group.stats, **group.custom}
                metric_types = {**self.types_map, **group.custom_schemas}

                for metric_field, metric_value in metric_record.items():
                    
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metric - {metrics_set.name}:{group_name}:{metric_field}')
                    
                    update_type = metric_types.get(metric_field)
                    update_function = self._update_map.get(update_type)
                    
                    update_function(
                        f'{metrics_set.name}_{group_name}_{metric_field}',
                        metric_value
                    )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to {self.statsd_type}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to {self.statsd_type}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            for custom_metric_name, custom_metric in metrics_set.custom_metrics.items():

                metric_type = self.stat_type_map.get(
                    custom_metric.metric_type,
                    'gauge'
                )

                update_function = self._update_map.get(metric_type)
                update_function(
                    f'{metrics_set.name}_{custom_metric_name}',
                    custom_metric.metric_value
                )


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to {self.statsd_type}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to {self.statsd_type}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                error_message = re.sub(
                    '[^0-9a-zA-Z]+', 
                    '_',
                    error.get(
                        'message'
                    ).lower()
                )

                update_function = self._update_map.get('count')
                update_function(f'{metrics_set.name}_errors_{error_message}', error.get('count'))


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to {self.statsd_type}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connection to {self.statsd_type} at - {self.host}:{self.port}')
        await self.connection.close()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connection to {self.statsd_type} at - {self.host}:{self.port}')


import uuid
import inspect
from typing import (
    List, 
    Callable, 
    Any, 
    Awaitable
)
from hedra.core.hooks.types.base.simple_context import SimpleContext
from hedra.core.hooks.types.base.hook_type import HookType


class Hook:

    def __init__(
        self, 
        name: str, 
        shortname: str,
        call: Callable[..., Awaitable[Any]], 
        stage: str = None,
        order: int=None,
        skip: bool=False,
        hook_type: HookType=None
    ) -> None:
        self.hook_id = str(uuid.uuid4())
        self.name = name
        self.shortname = shortname

        self._call: Callable[..., Awaitable[Any]] = call
        self.stage = stage
        self.hook_type = hook_type
        self.stage_instance: Any = None
        self.conditions: List[Callable[..., Any]] = []

        if call is None:
            call = self._default_call

        self.args = inspect.signature(call)
        self.params = self.args.parameters
        self.context: SimpleContext = SimpleContext()
        self.order: int = order
        self.skip = skip
        

    async def call(self, **kwargs):

        hook_args = {name: value for name, value in kwargs.items() if name in self.params}
        execute = await self._execute_call(**hook_args)

        if execute:
            result = await self._call(**hook_args)

            if isinstance(result, dict):
                return result

            return {
                **kwargs,
                self.name: result
            }


    async def _execute_call(self, **hook_args):
        execute = True
        for condition in self.conditions:
            execute = await condition(**{name: value for name, value in hook_args.items() if name in self.params})

        return execute
    
    async def _default_call(self, **kwargs):
        pass

    def copy(self):
        return Hook(
            self.name,
            self.shortname,
            self._call,
            stage=self.stage,
            order=self.order,
            skip=self.skip,
            hook_type=self.hook_type
        )



import asyncio
import sys
import socket
import struct
import time
import functools
import uuid
from typing import Tuple, Any
from .ping_type import PingType

if sys.platform.startswith("win"):
    if sys.version_info[0] > 3 or (sys.version_info[0] == 3 and sys.version_info[1] >= 8):
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

# ICMP types, see rfc792 for v4, rfc4443 for v6
ICMP_ECHO_REQUEST = 8
ICMP6_ECHO_REQUEST = 128
ICMP_ECHO_REPLY = 0
ICMP6_ECHO_REPLY = 129


class PingConnection:

    def __init__(self) -> None:
        self.proto_icmp = socket.getprotobyname("icmp")
        self.proto_icmp6 = socket.getprotobyname("ipv6-icmp")


    def checksum(self, buffer: bytes) -> bytes:
        """
        I'm not too confident that this is right but testing seems
        to suggest that it gives the same answers as in_cksum in ping.c
        :param buffer:
        :return:
        """
        sum = 0
        count_to = (len(buffer) / 2) * 2
        count = 0

        while count < count_to:
            this_val = buffer[count + 1] * 256 + buffer[count]
            sum += this_val
            sum &= 0xffffffff # Necessary?
            count += 2

        if count_to < len(buffer):
            sum += buffer[len(buffer) - 1]
            sum &= 0xffffffff # Necessary?

        sum = (sum >> 16) + (sum & 0xffff)
        sum += sum >> 16
        answer = ~sum
        answer &= 0xffff

        # Swap bytes. Bugger me if I know why.
        answer = answer >> 8 | (answer << 8 & 0xff00)

        return answer

    async def receive(
        self,
        active_socket: socket.socket, 
        id_: bytes, 
        timeout: float
    ) -> None:
        """
        receive the ping from the socket.
        :param my_socket:
        :param id_:
        :param timeout:
        :return:
        """
        loop = asyncio.get_event_loop()

        try:
            start = time.monotonic()
            elapsed = 0

            while elapsed < timeout:
                rec_packet = await loop.sock_recv(active_socket, 1024)
                time_received = time.perf_counter()

                if active_socket.family == socket.AddressFamily.AF_INET:
                    offset = 20
                else:
                    offset = 0

                icmp_header = rec_packet[offset:offset + 8]

                type, code, checksum, packet_id, sequence = struct.unpack(
                    "bbHHh", icmp_header
                )

                if type != ICMP_ECHO_REPLY and type != ICMP6_ECHO_REPLY:
                    continue

                if packet_id == id_:
                    data = rec_packet[offset + 8:offset + 8 + struct.calcsize("d")]
                    time_sent = struct.unpack("d", data)[0]

                    return time_received - time_sent
                
                elapsed = time.monotonic() - start

        except asyncio.TimeoutError:
            asyncio.get_event_loop().remove_writer(active_socket)
            asyncio.get_event_loop().remove_reader(active_socket)
            active_socket.close()

            raise TimeoutError("Ping timeout")


    def sendto_ready(
        self, 
        packet: bytes, 
        socket: socket.socket, 
        future: asyncio.Future, 
        dest: Tuple[str, ...]
    ):
        try:
            socket.sendto(packet, dest)
        except (BlockingIOError, InterruptedError):
            return  # The callback will be retried
        except Exception as exc:
            asyncio.get_event_loop().remove_writer(socket)
            future.set_exception(exc)
        else:
            asyncio.get_event_loop().remove_writer(socket)
            future.set_result(None)


    async def send(
        self,
        active_socket: socket.socket, 
        dest_addr: Tuple[str, ...], 
        id_: bytes, 
        family: int
    ):
        """
        Send one ping to the given >dest_addr<.
        :param my_socket:
        :param dest_addr:
        :param id_:
        :param timeout:
        :return:
        """
        icmp_type = ICMP_ECHO_REQUEST if family == socket.AddressFamily.AF_INET\
            else ICMP6_ECHO_REQUEST

        # Header is type (8), code (8), checksum (16), id (16), sequence (16)
        ping_checksum = 0

        # Make a dummy header with a 0 checksum.
        header = struct.pack(
            "BbHHh", 
            icmp_type, 
            0, 
            ping_checksum, 
            id_, 
            1
        )

        bytes_in_double = struct.calcsize("d")
        data = (192 - bytes_in_double) * "Q"
        data = struct.pack(
            "d", 
            time.perf_counter()
        ) + data.encode("ascii")

        # Calculate the checksum on the data and the dummy header.
        ping_checksum = self.checksum(header + data)

        # Now that we have the right checksum, we put that in. It's just easier
        # to make up a new header than to stuff it into the dummy.
        header = struct.pack(
            "BbHHh", 
            icmp_type, 
            0, 
            socket.htons(ping_checksum), 
            id_, 
            1
        )
        packet = header + data

        future = asyncio.get_event_loop().create_future()

        callback = functools.partial(
            self.sendto_ready, 
            packet=packet, 
            socket=active_socket, 
            dest=dest_addr, 
            future=future
        )

        asyncio.get_event_loop().add_writer(
            active_socket, 
            callback
        )

        await future


    async def ping(
        self,
        socket_config: Tuple[Any, ...]=None, 
        ping_type: PingType=PingType.ICMP,
        timeout: int=10
    ):
        """
        Returns either the delay (in seconds) or raises an exception.
        :param dest_addr:
        :param timeout:
        :param family:
        """

        family, type_, proto, _, address = socket_config

        if ping_type == PingType.ICMP:

            if family == socket.AddressFamily.AF_INET:
                proto = self.proto_icmp
            else:
                proto = self.proto_icmp6

        try:
            my_socket = socket.socket(family=family, type=type_, proto=proto)
            my_socket.setblocking(False)

        except OSError as e:
            msg = e.strerror

            if e.errno == 1:
                # Operation not permitted
                msg += (
                    " - Note that ICMP messages can only be sent from processes"
                    " running as root."
                )

                raise OSError(msg)

            raise

        my_id = uuid.uuid4().int & 0xFFFF

        if ping_type == PingType.ICMP:
            await self.send(my_socket, address, my_id, family)
            response = await self.receive(my_socket, my_id, timeout)

        my_socket.close()

        return response




import zstandard
from base64 import (
    b64encode,
    b64decode
)
from http.cookies import (
    BaseCookie,
    SimpleCookie
)
from hedra.distributed.env import Env, load_env
from hedra.distributed.encryption import AESGCMFernet
from hedra.distributed.middleware.base import (
    Middleware
)
from hedra.distributed.models.http import (
    Request,
    Response
)
from secrets import (
    compare_digest,
    token_urlsafe
)
from typing import (
    Tuple, 
    Optional, 
    List, 
    Set, 
    Dict, 
    Literal
)


class CRSF(Middleware):

    def __init__(
        self,
        secret_bytes_size: Optional[int]=16,
        required_paths: Optional[List[str]] = None,
        exempt_paths: Optional[List[str]] = None,
        sensitive_cookies: Optional[Set[str]] = None,
        safe_methods: List[
            Literal[
                "GET",
                "HEAD",
                "OPTIONS",
                "POST",
                "PUT",
                "PATCH",
                "DELETE",
                "TRACE"
            ]
        ] = [
            "GET", 
            "HEAD", 
            "OPTIONS", 
            "TRACE"
        ],
        cookie_name: str = "csrftoken",
        cookie_path: str = "/",
        cookie_domain: Optional[str] = None,
        cookie_secure: bool = False,
        cookie_httponly: bool = False,
        cookie_samesite: str = "lax",
        header_name: str = "x-csrftoken",
    ) -> None:
        
        env = load_env(Env)

        self.encryptor = AESGCMFernet(env)
        self.secret_bytes_size = secret_bytes_size

        self.required_paths = required_paths
        self.exempt_paths = exempt_paths
        self.sensitive_cookies = sensitive_cookies
        self.safe_methods = safe_methods
        self.cookie_name = cookie_name
        self.cookie_path = cookie_path
        self.cookie_domain = cookie_domain
        self.cookie_secure = cookie_secure
        self.cookie_httponly = cookie_httponly
        self.cookie_samesite = cookie_samesite
        self.header_name = header_name

        self._compressor = zstandard.ZstdCompressor()
        self._decompressor = zstandard.ZstdDecompressor()

        super().__init__(
            self.__class__.__name__,
            response_headers={}
        )

    async def __run__(
        self, 
        request: Request,
        response: Response,
        status: int
    ) -> Tuple[
        Tuple[Response, int],
        bool
    ]:

        crsf_cookie = request.cookies.get(self.cookie_name)

        request_path = request.path

        is_unsafe_method = request.method not in self.safe_methods

        path_is_required = False
        if self.required_paths:
            path_is_required = self._path_is_required(request_path)

        path_is_exempt = False
        if self.exempt_paths:
            path_is_exempt = self._path_is_exempt(request_path)
        
        has_sensitive_cookies = False
        if self.sensitive_cookies:
            has_sensitive_cookies = self._has_sensitive_cookies(request.cookies)

        is_sensitive = is_unsafe_method and not path_is_exempt and has_sensitive_cookies

        if path_is_required or is_sensitive:
            submitted_csrf_token = request.headers.get(self.header_name)

            csrf_tokens_match = False

            try:

                decoded_crsf_cookie: str = self.encryptor.decrypt(
                    self._decompressor.decompress(
                        b64decode(crsf_cookie.encode())
                    )
                )
                decoded_crsf_token: str = self.encryptor.decrypt(
                    self._decompressor.decompress(
                        b64decode(submitted_csrf_token.encode())
                    )
                )

                csrf_tokens_match = compare_digest(
                    decoded_crsf_cookie, 
                    decoded_crsf_token
                )

            except Exception:
                csrf_tokens_match = False

            
            crsf_match_failed = crsf_cookie is None or submitted_csrf_token is None or csrf_tokens_match is False

            if crsf_match_failed:
                return (
                    Response(
                        request.path,
                        request.method,
                        data="CSRF token verification failed"
                    ), 
                    403
                ), False

        crsf_cookie = request.cookies.get(self.cookie_name)

        response_headers = {}

        if crsf_cookie is None:

            cookie: BaseCookie = SimpleCookie()
            cookie_name = self.cookie_name

            crsf_token = self.encryptor.encrypt(
                token_urlsafe(
                    nbytes=self.secret_bytes_size
                ).encode()
            )

            cookie[cookie_name] = b64encode(
                self._compressor.compress(
                    crsf_token
                )
            ).decode()

            cookie[cookie_name]["path"] = self.cookie_path
            cookie[cookie_name]["secure"] = self.cookie_secure
            cookie[cookie_name]["httponly"] = self.cookie_httponly
            cookie[cookie_name]["samesite"] = self.cookie_samesite

            if self.cookie_domain is not None:
                cookie[cookie_name]["domain"] = self.cookie_domain  # pragma: no cover

            response_headers["set-cookie"] = cookie.output(header="").strip()

        if response and status:
            response.headers.update(response_headers)

            return (
                response,
                status
            ), True
        
        return (
            Response(
                request.path,
                request.method,
                headers=response_headers,
                data=None
            ), 
            200
        ), True
        

    def _has_sensitive_cookies(self, cookies: Dict[str, str]) -> bool:
        
        for sensitive_cookie in self.sensitive_cookies:
            if cookies.get(sensitive_cookie) is not None:
                return True
            
        return False

    def _path_is_required(self, path: str) -> bool:

        for required_url in self.required_paths:
            if required_url in path:
                return True
            
        return False

    def _path_is_exempt(self, path: str) -> bool:
        
        for exempt_path in self.exempt_paths:
            if exempt_path in path:
                return True
            
        return False


# -*- coding: utf-8 -*-
"""
h2/windows
~~~~~~~~~~

Defines tools for managing HTTP/2 flow control windows.

The objects defined in this module are used to automatically manage HTTP/2
flow control windows. Specifically, they keep track of what the size of the
window is, how much data has been consumed from that window, and how much data
the user has already used. It then implements a basic algorithm that attempts
to manage the flow control window without user input, trying to ensure that it
does not emit too many WINDOW_UPDATE frames.
"""
from __future__ import division



# The largest acceptable value for a HTTP/2 flow control window.
LARGEST_FLOW_CONTROL_WINDOW = 2**31 - 1


class WindowManager:

    __slots__ = (
        'max_window_size',
        'current_window_size',
        '_bytes_processed'
    )

    """
    A basic HTTP/2 window manager.

    :param max_window_size: The maximum size of the flow control window.
    :type max_window_size: ``int``
    """
    def __init__(self, max_window_size):
        self.max_window_size = max_window_size
        self.current_window_size = max_window_size
        self._bytes_processed = 0

    def window_consumed(self, size):
        """
        We have received a certain number of bytes from the remote peer. This
        necessarily shrinks the flow control window!

        :param size: The number of flow controlled bytes we received from the
            remote peer.
        :type size: ``int``
        :returns: Nothing.
        :rtype: ``None``
        """
        self.current_window_size -= size

    def window_opened(self, size):
        """
        The flow control window has been incremented, either because of manual
        flow control management or because of the user changing the flow
        control settings. This can have the effect of increasing what we
        consider to be the "maximum" flow control window size.

        This does not increase our view of how many bytes have been processed,
        only of how much space is in the window.

        :param size: The increment to the flow control window we received.
        :type size: ``int``
        :returns: Nothing
        :rtype: ``None``
        """
        self.current_window_size += size

        if self.current_window_size > self.max_window_size:
            self.max_window_size = self.current_window_size

    def process_bytes(self, size):
        """
        The application has informed us that it has processed a certain number
        of bytes. This may cause us to want to emit a window update frame. If
        we do want to emit a window update frame, this method will return the
        number of bytes that we should increment the window by.

        :param size: The number of flow controlled bytes that the application
            has processed.
        :type size: ``int``
        :returns: The number of bytes to increment the flow control window by,
            or ``None``.
        :rtype: ``int`` or ``None``
        """

        if size is None:
            size = 0
        
        self._bytes_processed += size
        if not self._bytes_processed:
            return None

        max_increment = (self.max_window_size - self.current_window_size)
        increment = 0

        min_threshold =  (self.current_window_size == 0) and (self._bytes_processed > min(1024, self.max_window_size // 4))
        max_threshold = self._bytes_processed >= (self.max_window_size // 2)

        # Note that, even though we may increment less than _bytes_processed,
        # we still want to set it to zero whenever we emit an increment. This
        # is because we'll always increment up to the maximum we can.
        if min_threshold or max_threshold:
            increment = min(self._bytes_processed, max_increment)
            self._bytes_processed = 0

        self.current_window_size += increment
        return increment


from __future__ import annotations
from typing import Dict, Union
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.http.result import HTTPResult
from .action import GraphQLAction


class GraphQLResult(HTTPResult):

    def __init__(self, action: GraphQLAction, error: Exception = None) -> None:
        super().__init__(action, error)

        self.type = RequestTypes.GRAPHQL



from hedra.data.connectors.common.connector_type import ConnectorType
from pydantic import (
    BaseModel,
    StrictStr
)



class MySQLConnectorConfig(BaseModel):
    host: StrictStr='127.0.0.1'
    database: StrictStr
    username: StrictStr
    password: StrictStr
    table_name: StrictStr
    connector_type: ConnectorType=ConnectorType.MySQL

    class Config:
        arbitrary_types_allowed = True


from .cassandra_graphql_http2_result_schema import CassandraGraphQLHTTP2ResultSchema
from .cassandra_graphql_result_schema import CassandraGraphQLResultSchema
from .cassandra_grpc_result_schema import CassandraGRPCResultSchema
from .cassandra_http_result_schema import CassandraHTTPResultSchema
from .cassandra_http2_result_schema import CassandraHTTP2ResultSchema
from .cassandra_http3_result_schema import CassandraHTTP3ResultSchema
from .cassandra_playwright_result_schema import CassandraPlaywrightResultSchema
from .cassandra_task_result_schema import CassandraTaskResultSchema
from .cassandra_udp_result_schema import CassandraUDPResultSchema
from .cassandra_websocket_result_schema import CassandraWebsocketResultSchema


import json
from typing import (
    Dict, 
    Iterator, 
    Union, 
    List, 
    Any, 
    Tuple
)
from urllib.parse import urlencode
from hedra.core.engines.types.common.hooks import Hooks
from hedra.core.engines.types.common.base_action import BaseAction
from hedra.core.engines.types.common.constants import NEW_LINE
from hedra.core.engines.types.common.protocols.shared.writer import Writer
from hedra.core.engines.types.common import URL
from hedra.core.engines.types.common.types import RequestTypes


class HTTPAction(BaseAction):

    __slots__ = (
        'action_id',
        'method',
        'listeners',
        'type',
        'url',
        'protocols',
        '_headers',
        '_data',
        'encoded_data',
        'encoded_headers',
        'is_stream',
        'ssl_context',
        'redirects',
        'action_args',
        'mutations',
        '_header_items'
    )
    
    def __init__(
        self,
        name: str, 
        url: str, 
        method: str = 'GET', 
        headers: Dict[str, str] = {}, 
        data: Union[str, dict, Iterator, bytes, None] = None, 
        user: str=None, 
        tags: List[Dict[str, str]] = [],
        redirects: int=3
    ) -> None:
        super(HTTPAction, self).__init__(
            name,
            user,
            tags
        )

        self.method = method.upper()
        self.type = RequestTypes.HTTP

        address_family, protocol = self.protocols[self.type]
        self.url = URL(url, family=address_family, protocol=protocol)

        self._headers = headers
        self._header_items: List[Tuple[str, str]] = list(headers.items())
        self._data = data

        self.encoded_data = None
        self.encoded_headers = None
        self.is_stream = False
        self.ssl_context = None
        self.redirects = redirects
        self.hooks: Hooks[HTTPAction] = Hooks()
        self.action_args: Dict[str, Any] = {}

    @property
    def size(self):
        if self.encoded_data:
            return len(self.encoded_data)

        else:
            return 0

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = value
        self.encoded_data = None

    @property
    def headers(self):
        return self._headers

    @headers.setter
    def headers(self, value: Dict[str, str]):
        self._headers = value
        self._header_items: List[Tuple[str, str]] = list(value.items())
        self.encoded_headers = None

    def setup(self):

        if self.encoded_data is None:
            self._setup_data()

        if self.encoded_headers is None:
            self._setup_headers()

    def _setup_data(self):
        if self._data:
            if isinstance(self._data, Iterator):
                chunks = []
                for chunk in self._data:
                    chunk_size = hex(len(chunk)).replace("0x", "") + NEW_LINE
                    encoded_chunk = chunk_size.encode() + chunk + NEW_LINE.encode()
                    self.size += len(encoded_chunk)
                    chunks.append(encoded_chunk)

                self.is_stream = True
                self.encoded_data = chunks

            else:

                if isinstance(self._data, dict):
                    self.encoded_data = json.dumps(
                        self._data
                    ).encode()

                elif isinstance(self._data, tuple):
                    self.encoded_data = urlencode(
                        self._data
                    ).encode()

                elif isinstance(self._data, str):
                    self.encoded_data = self._data.encode()

    def _setup_headers(self) -> Union[bytes, Dict[str, str]]:
    
        get_base = f"{self.method} {self.url.path} HTTP/1.1{NEW_LINE}"

        port = self.url.port or (443 if self.url.scheme == "https" else 80)

        hostname = self.url.parsed.hostname.encode("idna").decode()

        if port not in [80, 443]:
            hostname = f'{hostname}:{port}'

        header_items = [
            ("HOST", hostname),
            ("User-Agent", "mercury-http"),
            ("Keep-Alive", "timeout=60, max=100000"),
            ("Content-Length", self.size)
        ]

        header_items.extend(self._header_items)

        for key, value in header_items:
            get_base += f"{key}: {value}{NEW_LINE}"

        self.encoded_headers = (get_base + NEW_LINE).encode()

    def write_chunks(self, writer: Writer):
        for chunk in self.data:
            writer.write(chunk)

        writer.write(("0" + NEW_LINE * 2).encode())



__all__ = ('Lock', 'Event', 'Condition', 'Semaphore', 'BoundedSemaphore')

import collections

from asyncio import (
    exceptions,
    mixins,
)


class BatchedSemaphore(mixins._LoopBoundMixin):
    """A Semaphore implementation.

    A semaphore manages an internal counter which is decremented by each
    acquire() call and incremented by each release() call. The counter
    can never go below zero; when acquire() finds that it is zero, it blocks,
    waiting until some other thread calls release().

    Semaphores also support the context management protocol.

    The optional argument gives the initial value for the internal
    counter; it defaults to 1. If the value given is less than 0,
    ValueError is raised.
    """

    def __init__(self, value=1, *, loop=None):
        super().__init__()
        if value < 0:
            raise ValueError("Semaphore initial value must be >= 0")
        self._value = value
        self._waiters = collections.deque()
        self._wakeup_scheduled = False

    def __repr__(self):
        res = super().__repr__()
        extra = 'locked' if self.locked() else f'unlocked, value:{self._value}'
        if self._waiters:
            extra = f'{extra}, waiters:{len(self._waiters)}'
        return f'<{res[1:-1]} [{extra}]>'

    def _wake_up_next(self):
        while self._waiters:
            waiter = self._waiters.popleft()
            if not waiter.done():
                waiter.set_result(None)
                self._wakeup_scheduled = True
                return

    def locked(self):
        """Returns True if semaphore can not be acquired immediately."""
        return self._value == 0

    async def acquire(self, workers_count: int):
        """Acquire a semaphore.

        If the internal counter is larger than zero on entry,
        decrement it by one and return True immediately.  If it is
        zero on entry, block, waiting until some other coroutine has
        called release() to make it larger than 0, and then return
        True.
        """
        # _wakeup_scheduled is set if *another* task is scheduled to wakeup
        # but its acquire() is not resumed yet
        while self._wakeup_scheduled or self._value <= 0:
            fut = self._get_loop().create_future()
            self._waiters.append(fut)
            try:
                await fut
                # reset _wakeup_scheduled *after* waiting for a future
                self._wakeup_scheduled = False
            except exceptions.CancelledError:
                self._wake_up_next()
                raise
        self._value -= workers_count
        return True

    def release(self, workers_count: int):
        """Release a semaphore, incrementing the internal counter by one.
        When it was zero on entry and another coroutine is waiting for it to
        become larger than zero again, wake up that coroutine.
        """
        self._value += workers_count
        self._wake_up_next()


import math
from hedra.distributed.models.http import Limit
from .base_limiter import BaseLimiter


class SlidingWindowLimiter(BaseLimiter):

    __slots__ = (
        "max_rate",
        "time_period",
        "_rate_per_sec",
        "_level",
        "_waiters",
        "_loop",
        "_current_time",
        "_previous_count"
    )

    def __init__(
        self, 
        limit: Limit
    ) -> None:

        super().__init__(
            limit.max_requests,
            limit.period,
            reject_requests=limit.reject_requests
        )

        self._current_time = self._loop.time()
        self._previous_count = limit.max_requests

    def has_capacity(self, amount: float = 1) -> bool:

        if (self._loop.time() - self._current_time) > self.time_period:
            self._current_time = math.floor(
                self._loop.time()/self.time_period
            ) * self.time_period

            self._previous_count = self._level
            self._level = 0

        self._rate_per_sec = (
            self._previous_count * (
                self.time_period - (self._loop.time() - self._current_time)
            )/self.time_period
        ) + (self._level + amount)

        if self._rate_per_sec < self.max_rate:

            for fut in self._waiters.values():
                if not fut.done():
                    fut.set_result(True)
                    break

        return self._rate_per_sec <= self.max_rate


from typing import Any, Tuple
from hedra.core.graphs.stages.base.stage import Stage
from pydantic import BaseModel, validator


class DependsValidator(BaseModel):
    stages: Tuple[Any, ...]

    class Config:
        arbitrary_types_allowed = True

    @validator('stages')
    def validate_stages(cls, vals):
        assert len(vals) > 0
        assert len(vals) == len(set(vals))

        for val in vals:
            assert issubclass(val, Stage)

        return vals



from typing import (
    List, 
    Union, 
    Dict, 
    Any, 
    Callable, 
    Awaitable, 
    Tuple,
)
from hedra.core.engines.types.common.base_action import BaseAction
from hedra.core.engines.types.common.base_engine import BaseEngine
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.base.hook_metadata import HookMetadata


class ActionHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        *names: Tuple[str, ...],
        weight: int=1, 
        order: int=1, 
        skip: bool=False,
        metadata: Dict[str, Union[str, int]]={}
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            skip=skip,
            order=order,
            hook_type=HookType.ACTION,
        )
        
        self.names = list(set(names))
        self.session: BaseEngine = None
        self.action: BaseAction = None
        self.checks: List[Any] = []
        self.before: List[str] = []
        self.after: List[str] = []
        self.is_notifier = False
        self.is_listener = False
        self.channels: List[Any] = []
        self.notifiers: List[Any] = []
        self.listeners: List[Any] = []
        self.metadata = HookMetadata(
            weight=weight,
            order=order,
            **metadata
        )

    def copy(self):
        action_hook = ActionHook(
            self.name,
            self.shortname,
            self._call,
            weight=self.metadata.weight,
            order=self.order,
            skip=self.skip,
            metadata={
                **self.metadata.copy()
            }
        )

        action_hook.checks = list(self.checks)
        action_hook.stage = self.stage

        return action_hook

    async def call(self, *args, **kwargs):
        return await self._call(*args, **kwargs)
    
    def to_dict(self) -> str:
        return {
            'name': self.name,
            'shortname': self.shortname,
            'skip': self.skip,
            'hook_type': HookType.ACTION,
            'names': self.names,
            'checks': [
                check if isinstance(check, str) else check.name for check in self.checks
            ],
            'channels': [
                channel if isinstance(channel, str) else channel.name for channel in self.channels
            ],
            'notifiers': [
                notifier if isinstance(notifier, str) else notifier.name for notifier in self.notifiers
            ],
            'listeners': [
                listener if isinstance(listener, str) else listener.name for listener in self.listeners
            ],
            'order': self.metadata.order,
            'weight': self.metadata.weight,
            'user': self.metadata.user,
            'tags': self.metadata.tags
        }



import functools
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.registrar import registrar
from .validator import ConditionHookValidator


@registrar(HookType.CONDITION)
def condition(
    *names,
    order: int=1,
    skip: bool=False
):
    ConditionHookValidator(
        names=names,
        order=order,
        skip=skip
    )

    def wrapper(func):

        @functools.wraps(func)
        def decorator(*args, **kwargs):
            return func(*args, **kwargs)

        return decorator

    return wrapper


import inspect
from inspect import Parameter
from hedra.plugins.types.common.plugin import Plugin
from hedra.plugins.types.common.plugin_hook import PluginHook
from hedra.plugins.types.common.types import PluginHooks
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.common.registrar import plugin_registrar
from typing import (
    Dict, 
    Any, 
    Union, 
    Mapping
)
from hedra.versioning.flags.types.unstable.flag import unstable_threadsafe
from .types import ExtensionType


class ExtensionPlugin(Plugin):
    type=PluginType.EXTENSION
    name: str=None
    
    def __init__(self) -> None:
        super(
            ExtensionPlugin
        ).__init__(self)

        self.hooks: Dict[PluginHooks, PluginHook] = {}
        self.name = self.name
        self._args: Dict[str, Mapping[str, Parameter]] = {}

        methods = inspect.getmembers(self, predicate=inspect.ismethod) 
        for _, method in methods:

                method_name = method.__qualname__
                hook: PluginHook = plugin_registrar.all.get(method_name)
                
                if hook:
                    hook.call = hook.call.__get__(self, self.__class__)
                    setattr(self, hook.shortname, hook.call)

                    self.hooks[hook.hook_type] = hook
                    args = inspect.signature(hook.call)

                    self._args[hook.hook_type] = args.parameters

        self.extension_type: ExtensionType = None
        unstable_threadsafe()

    async def execute(
        self,
        **kwargs
    ) -> Dict[str, Any]:
        next_args = {
             **kwargs
        }

        execute_hook = self.hooks.get(PluginHooks.ON_EXTENSION_PREPARE)
        hook_args = self._args.get(PluginHooks.ON_EXTENSION_PREPARE)

        result: Union[Dict[str, Any], Any] = await execute_hook.call(
            **{
                 name: value for name, value in next_args.items() if name in hook_args
            }
        )

        if isinstance(result, dict) is False:
             result = {
                  'extension_data': result
             }

        next_args = {
             **kwargs,
             **result
        }

        prepare_hook = self.hooks.get(PluginHooks.ON_EXTENSION_EXECUTE)
        hook_args = self._args.get(PluginHooks.ON_EXTENSION_EXECUTE)
        
        return await prepare_hook.call(
             **{
                 name: value for name, value in next_args.items() if name in hook_args
            }
        )


from hedra.core.engines.types.common.types import RequestTypes
from .types import (
    GraphQLProcessedResult,
    GraphQLHTTP2ProcessedResult,
    GRPCProcessedResult,
    HTTPProcessedResult,
    HTTP2ProcessedResult,
    HTTP3ProcessedResult,
    PlaywrightProcessedResult,
    TaskProcessedResult,
    UDPProcessedResult,
    WebsocketProcessedResult,
)



results_types = {
    RequestTypes.GRAPHQL: GraphQLProcessedResult,
    RequestTypes.GRAPHQL_HTTP2: GraphQLHTTP2ProcessedResult,
    RequestTypes.GRPC: GRPCProcessedResult,
    RequestTypes.HTTP: HTTPProcessedResult,
    RequestTypes.HTTP2: HTTP2ProcessedResult,
    RequestTypes.HTTP3: HTTP3ProcessedResult,
    RequestTypes.PLAYWRIGHT: PlaywrightProcessedResult,
    RequestTypes.TASK: TaskProcessedResult,
    RequestTypes.UDP: UDPProcessedResult,
    RequestTypes.WEBSOCKET: WebsocketProcessedResult
}




from __future__ import annotations
from typing import Dict, Union
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.core.engines.types.common.types import RequestTypes
from .command import PlaywrightCommand, Page, URL


class PlaywrightResult(BaseResult):

    __slots__ = (
        'action_id',
        'url',
        'headers',
        'command',
        'selector',
        'attribute',
        'x_coord',
        'y_coord',
        'frame',
        'key',
        'text',
        'expression',
        'args',
        'filepath',
        'file',
        'option',
        'event',
        'is_checked'
    )

    def __init__(self, command: PlaywrightCommand, error: Exception=None) -> None:
        super(
            PlaywrightResult,
            self
        ).__init__(
            command.action_id,
            command.name,
            command.url.location,
            command.metadata.user,
            command.metadata.tags,
            RequestTypes.PLAYWRIGHT,
            error
        )

        self.url = command.url.location
        self.headers = command.url.headers
        self.command = command.command
        self.selector = command.page.selector
        self.attribute = command.page.attribute
        self.x_coord = command.page.x_coordinate
        self.y_coord = command.page.y_coordinate
        self.frame = command.page.frame
        self.key = command.input.key
        self.text = command.input.text
        self.expression = command.input.expression
        self.args = command.input.args
        self.filepath = command.input.filepath
        self.file = command.input.file
        self.option = command.input.option
        self.event = command.options.event
        self.timeout = command.options.timeout
        self.is_checked = command.options.is_checked



from ctypes import Union
import inspect
import math
from typing import Any, Dict
from hedra.core.graphs.stages.optimize.optimization.algorithms.types.base_algorithm import BaseAlgorithm
from hedra.plugins.types.common.plugin_hook import PluginHook
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.common.types import PluginHooks
from hedra.plugins.types.common.plugin import Plugin
from hedra.plugins.types.common.registrar import plugin_registrar


class OptimizerPlugin(BaseAlgorithm, Plugin):
    type=PluginType.OPTIMIZER
    name: str=None

    def __init__(self, config: Dict[str, Any]) -> None:
        self.hooks: Dict[PluginHooks, PluginHook] = {}
        
        methods = inspect.getmembers(self, predicate=inspect.ismethod) 
        for _, method in methods:

                method_name = method.__qualname__
                hook: PluginHook = plugin_registrar.all.get(method_name)
                
                if hook:
                    hook.call = hook.call.__get__(self, self.__class__)
                    setattr(self, hook.shortname, hook.call)

                    self.hooks[hook.hook_type] = hook


        on_get_params = self.hooks.get(PluginHooks.ON_OPTIMIZER_GET_PARAMS)
        on_update_params = self.hooks.get(PluginHooks.ON_OPTIMIZER_UPDATE_PARAMS)
        on_optimize = self.hooks.get(PluginHooks.ON_OPTIMIZE)
        self.name = self.name

        if on_get_params:
            self.get_params = on_get_params.call

        if on_update_params:
            self.update_params = on_update_params.call

        self.optimize = on_optimize.call

        super().__init__(config)




        
        



from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictFloat,
    StrictBool,
    AnyHttpUrl
)


from typing import List, Dict, Optional, Union


class GraphQLHTTP2ActionTag(BaseModel):
    name: StrictStr
    value: StrictStr


class GraphQLHTTP2ActionValidator(BaseModel):
    engine: StrictStr
    name: StrictStr
    url: AnyHttpUrl
    method: StrictStr='GET'
    headers: Dict[StrictStr, StrictStr]={}
    query: StrictStr
    operation_name: StrictStr
    variables: Dict[str, Union[StrictStr, StrictInt, StrictFloat, StrictBool, None]]
    weight: Optional[Union[StrictInt, StrictFloat]]
    order: Optional[StrictInt]
    user: Optional[StrictStr]
    tags: List[GraphQLHTTP2ActionTag]=[]


from hedra.distributed.models.base.message import Message
from pydantic import (
    StrictStr,
    StrictInt
)
from typing import List, Optional, Tuple
from .vote_result import VoteResult
from .healthcheck import HealthStatus
from .logs import Entry, NodeState


class RaftMessage(Message):
    source_host: StrictStr
    source_port: StrictInt
    elected_leader: Optional[Tuple[StrictStr, StrictInt]]
    failed_node: Optional[Tuple[StrictStr, StrictInt]]
    vote_result: Optional[VoteResult]
    raft_node_status: NodeState
    status: HealthStatus
    entries: Optional[List[Entry]]
    term_number: StrictInt
    received_timestamp: Optional[StrictInt]



from pydantic import (
    BaseModel, 
    StrictFloat, 
    StrictStr,
    validator
)
from typing import Tuple
from .mutation_type import MutationType


class MutationValidator(BaseModel):
    name: StrictStr
    chance: StrictFloat
    targets: Tuple[StrictStr, ...]
    mutation_type: MutationType

    class Config:
        arbitrary_types_allowed=True

    @validator('targets')
    def validate_targets(cls, val):
        assert len(val) > 0, "Mutations must target more than one Action or Task hook."
        return val


from hedra.versioning.flags.types.base.flag_type import FlagTypes
from hedra.versioning.flags.types.base.registry import flag_registrar


@flag_registrar(FlagTypes.UNSAFE_FEATURE)
def unsafe(feature):

    def wrap_feature(*args, **kwargs):

        return feature(*args, **kwargs)
    
    return wrap_feature


from hedra.data.connectors.aws_lambda.aws_lambda_connector_config import AWSLambdaConnectorConfig
from hedra.data.connectors.bigtable.bigtable_connector_config import BigTableConnectorConfig
from typing import Tuple, Optional, Union


from pydantic import (
    BaseModel, 
    StrictStr, 
    StrictInt,
    StrictBool
)


class LoadHookValidator(BaseModel):
    names: Optional[Tuple[StrictStr, ...]]
    loader: Union[
        AWSLambdaConnectorConfig,
        BigTableConnectorConfig,
    ]
    order: StrictInt
    skip: StrictBool

    class Config:
        arbitrary_types_allowed=True





import asyncio
import uuid
import signal
import psutil
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from typing import List, Dict
from hedra.logging import HedraLogger
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import (
    MetricsSet
)
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet, 
    SystemMetricsCollection,
    SessionMetricsCollection
)
from .bigtable_config import BigTableConfig

try:
    from google.cloud import bigtable
    from google.auth import load_credentials_from_file
    has_connector = True

except Exception:
    bigtable = None
    Credentials = None
    has_connector = False


def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop
): 
    try:
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass
    


class BigTable:

    def __init__(self, config: BigTableConfig) -> None:

        self.service_account_json_path = config.service_account_json_path

        self.instance_id = config.instance_id
        self.events_table_id = config.events_table
        self.metrics_table_id = config.metrics_table
        self.streams_table_id = config.streams_table

        self.experiments_table_id = config.experiments_table
        self.variants_table_id = f'{config.experiments_table}_variants'
        self.mutations_table_id = f'{config.experiments_table}_mutations'

        self.shared_metrics_table_id = f'{self.metrics_table_id}_metrics'
        self.custom_metrics_table_ids = {}
        self.errors_table_id = f'{self.metrics_table_id}_errors'

        self.session_system_metrics_table_id = f'{config.system_metrics_table}_session'
        self.stage_system_metrics_table_id = f'{config.system_metrics_table}_stage'

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))

        self._events_column_family_id = f'{self.events_table_id}_columns'
        self._metrics_column_family_id = f'{self.metrics_table_id}_columns'
        self._streams_column_family_id = f'{self.streams_table_id}_columns'

        self._experiments_column_family_id = f'{self.experiments_table_id}_columns'
        self._variants_column_family_id = f'{self.experiments_table_id}_variants_columns'
        self._mutations_column_family_id = f'{self.experiments_table_id}_mutations_columns'


        self.session_system_metrics_column_family_id= f'{config.system_metrics_table}_session_columns'
        self.stage_system_metrics_column_family_id = f'{config.system_metrics_table}_stage_columns'

        self._shared_metrics_column_family_id = f'{self.metrics_table_id}_shared_columns'
        self._custom_metrics_column_family_id = f'{self.metrics_table_id}_custom_columns'
        self._errors_column_family_id = f'{self.metrics_table_id}_errors_columns'

        self.instance = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self._events_table = None
        self._stage_metrics_table = None
        self._streams_table = None

        self._experiments_table = None
        self._variants_table = None
        self._mutations_table = None

        self._custom_metrics_tables = {}
        self._metrics_table = None
        self._errors_table = None

        self._session_system_metrics_table = None
        self._stage_system_metrics_table = None

        self._experiments_table_columns = None
        self._variants_table_columns = None
        self._mutations_table_columns = None

        self._events_table_columns = None
        self._stage_metrics_table_columns = None
        self._streams_table_columns = None

        self._session_system_metrics_columns = None
        self._stage_system_metrics_columns = None

        self._custom_metrics_table_columns = {}
        self._metrics_table_columns = None
        self._errors_table_columns = None

        self.credentials = None
        self.client = None
        self._loop = asyncio.get_event_loop()

    async def connect(self):

        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            self._loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(
                    signame,
                    self._executor,
                    self._loop
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening amd authorizing connection to Google Cloud - Loading account config from - {self.service_account_json_path}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Opening session - {self.session_uuid}')

        credentials, project_id = load_credentials_from_file(self.service_account_json_path)
        self.client = bigtable.Client(
            project=project_id,
            credentials=credentials,
            admin=True
        )
        self.instance = self.client.instance(self.instance_id)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opened connection to Google Cloud - Created Client Instance - ID:{self.instance_id}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opened connection to Google Cloud - Loaded account config from - {self.service_account_json_path}')

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to Table {self.session_system_metrics_table_id}')
        self._session_system_metrics_table = self.instance.table(self.session_system_metrics_table_id)
        
        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Session System Metrics Table - {self.session_system_metrics_table_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._session_system_metrics_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Session System Metrics Table - {self.session_system_metrics_table_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Session System Metrics Column Family - {self.session_system_metrics_table_id} - if not exists')

        self._session_system_metrics_columns = self._session_system_metrics_table.column_family(
            self.session_system_metrics_column_family_id
        )

        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Session System Metrics Column for Column Family - {self.session_system_metrics_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._session_system_metrics_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Session System Metrics Column for Column Family - {self.session_system_metrics_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Session System Metrics Column for Column Family - {self.session_system_metrics_column_family_id} - if not exists')

        metrics_sets: List[SessionMetricsCollection] = []

        for metrics_set in system_metrics_sets:
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                metrics_sets.append(monitor_metrics)
                
            for monitor_metrics in metrics_set.session_memory_metrics.values():
                metrics_sets.append(monitor_metrics)


        rows = []
        for metrics_set in metrics_sets:

            row_key = f'{metrics_set.name}_{str(uuid.uuid4())}'
            row = self._variants_table.direct_row(row_key)

            for field, value in metrics_set.record:
                if not isinstance(value, bytes):
                    value = f'{value}'.encode()

                row.set_cell(
                    self._variants_column_family_id,
                    field,
                    value,
                    timestamp=datetime.now()
                )

            rows.append(row)

        await self._loop.run_in_executor(
            self._executor,
            self._variants_table.mutate_rows,
            rows
        )

    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to Table {self.stage_system_metrics_table_id}')
        self._stage_system_metrics_table = self.instance.table(self.stage_system_metrics_table_id)
        
        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Stage System Metrics Table - {self.stage_system_metrics_table_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._stage_system_metrics_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Stage System Metrics Table - {self.stage_system_metrics_table_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Stage System Metrics Column Family - {self.stage_system_metrics_table_id} - if not exists')

        self._stage_system_metrics_columns = self._stage_system_metrics_table.column_family(
            self.stage_system_metrics_column_family_id
        )

        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Stage System Metrics Column for Column Family - {self.stage_system_metrics_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._stage_system_metrics_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Stage System Metrics Column for Column Family - {self.stage_system_metrics_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Stage System Metrics Column for Column Family - {self.stage_system_metrics_column_family_id} - if not exists')

        metrics_sets: List[SystemMetricsCollection] = []

        for metrics_set in system_metrics_sets:

            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    metrics_sets.append(stage_mb_per_vu_metrics)

        rows = []
        for metrics_set in metrics_sets:

            row_key = f'{metrics_set.name}_{str(uuid.uuid4())}'
            row = self._variants_table.direct_row(row_key)

            for field, value in metrics_set.record:
                if not isinstance(value, bytes):
                    value = f'{value}'.encode()

                row.set_cell(
                    self._variants_column_family_id,
                    field,
                    value,
                    timestamp=datetime.now()
                )

            rows.append(row)

        await self._loop.run_in_executor(
            self._executor,
            self._variants_table.mutate_rows,
            rows
        )

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to Table {self.streams_table_id}')
        self._streams_table = self.instance.table(self.streams_table_id)
        
        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Streams Column Family - {self._streams_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._streams_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Streams Column Family - {self._streams_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Streams Column Family - {self._streams_column_family_id} - if not exists')

        self._streams_table_columns = self._streams_table.column_family(
            self._streams_column_family_id
        )

        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Streams Column for Column Family - {self._streams_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._streams_table_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Streams Column for Column Family - {self._streams_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Streams Column for Column Family - {self._streams_column_family_id} - if not exists')

        rows = []
        for stage_name, stream  in stream_metrics.items():
            for group_name, group in stream.grouped.items():
                stream_record = {
                    'name': f'{stage_name}_stream',
                    'stage': stage_name,
                    'group': group_name,
                    **group
                }

                row_key = f'{stage_name}_stream_{str(uuid.uuid4())}'
                row = self._variants_table.direct_row(row_key)

                for field, value in stream_record.items():
                    if not isinstance(value, bytes):
                        value = f'{value}'.encode()

                    row.set_cell(
                        self._variants_column_family_id,
                        field,
                        value,
                        timestamp=datetime.now()
                    )

                rows.append(row)

        await self._loop.run_in_executor(
            self._executor,
            self._variants_table.mutate_rows,
            rows
        )

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to Table {self.experiments_table_id}')
        self._experiments_table = self.instance.table(self.experiments_table_id)
        
        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Experiments Column Family - {self._experiments_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._experiments_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Experiments Column Family - {self._experiments_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Experiments Column Family - {self._experiments_column_family_id} - if not exists')

        self._experiments_table_columns = self._experiments_table.column_family(
            self._experiments_column_family_id
        )

        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Experiments Column for Column Family - {self._experiments_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._experiments_table_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Experiments Column for Column Family - {self._experiments_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Experiments Column for Column Family - {self._experiments_column_family_id} - if not exists')

        rows = []
        for experiment in experiment_metrics.experiment_summaries:

            row_key = f'{experiment.experiment_name}_{str(uuid.uuid4())}'
            row = self._experiments_table.direct_row(row_key)

            for field, value in experiment.record.items():
                if not isinstance(value, bytes):
                    value = f'{value}'.encode()

                row.set_cell(
                    self._experiments_column_family_id,
                    field,
                    value,
                    timestamp=datetime.now()
                )

            rows.append(row)

        await self._loop.run_in_executor(
            self._executor,
            self._experiments_table.mutate_rows,
            rows
        )

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to Table {self.variants_table_id}')
        self._variants_table = self.instance.table(self.variants_table_id)
        
        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Variants Column Family - {self._variants_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._variants_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Variants Column Family - {self._variants_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Variants Column Family - {self._variants_column_family_id} - if not exists')

        self._variants_table_columns = self._variants_table.column_family(
            self._variants_column_family_id
        )

        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Variants Column for Column Family - {self._variants_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._variants_table_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Variants Column for Column Family - {self._variants_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Variants Column for Column Family - {self._variants_column_family_id} - if not exists')

        rows = []
        for variant in experiment_metrics.variant_summaries:

            row_key = f'{variant.variant_name}_{str(uuid.uuid4())}'
            row = self._variants_table.direct_row(row_key)

            for field, value in variant.record:
                if not isinstance(value, bytes):
                    value = f'{value}'.encode()

                row.set_cell(
                    self._variants_column_family_id,
                    field,
                    value,
                    timestamp=datetime.now()
                )

            rows.append(row)

        await self._loop.run_in_executor(
            self._executor,
            self._variants_table.mutate_rows,
            rows
        )

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to Table {self.mutations_table_id}')
        self._mutations_table = self.instance.table(self.mutations_table_id)
        
        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Mutations Column Family - {self._mutations_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._mutations_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Mutations Column Family - {self._mutations_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Mutations Column Family - {self._mutations_column_family_id} - if not exists')

        self._mutations_table_columns = self._mutations_table.column_family(
            self._mutations_column_family_id
        )

        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Mutations Column for Column Family - {self._mutations_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._mutations_table_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Mutations Column for Column Family - {self._mutations_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Mutations Column for Column Family - {self._mutations_column_family_id} - if not exists')

        rows = []
        for mutation in experiment_metrics.mutation_summaries:

            row_key = f'{mutation.mutation_name}_{str(uuid.uuid4())}'
            row = self._mutations_table.direct_row(row_key)

            for field, value in mutation.record:
                if not isinstance(value, bytes):
                    value = f'{value}'.encode()

                row.set_cell(
                    self._mutations_column_family_id,
                    field,
                    value,
                    timestamp=datetime.now()
                )

            rows.append(row)

        await self._loop.run_in_executor(
            self._executor,
            self._mutations_table.mutate_rows,
            rows
        )

    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Table {self.events_table_id}')
        self._events_table = self.instance.table(self.events_table_id)
        
        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Events Column Family - {self._events_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._events_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Events Column Family - {self._events_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Events Column Family - {self._events_column_family_id} - if not exists')

        self._events_table_columns = self._events_table.column_family(
            self._events_column_family_id
        )

        try:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Events Column for Column Family - {self._events_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._events_table_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Events Column for Column Family - {self._events_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Events Column for Column Family - {self._events_column_family_id} - if not exists')

        rows = []
        for event in events:
            row_key = f'{event.name}_{str(uuid.uuid4())}'
            row = self._events_table.direct_row(row_key)

            for field, value in event.record.items():
                if not isinstance(value, bytes):
                    value = f'{value}'.encode()

                row.set_cell(
                    self._events_column_family_id,
                    field,
                    value,
                    timestamp=datetime.now()
                )

            rows.append(row)

        await self._loop.run_in_executor(
            self._executor,
            self._events_table.mutate_rows,
            rows
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Table {self.events_table_id}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Table {self.shared_metrics_table_id}')

        stage_metrics_table = self.instance.table(self.shared_metrics_table_id)
            
        try:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Shared Metrics Column Family - {self._shared_metrics_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                stage_metrics_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Shared Metrics Column Family - {self._shared_metrics_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Shared Metrics Column Family - {self._shared_metrics_column_family_id} - if not exists')

        self._metrics_table_columns = stage_metrics_table.column_family(
            self._shared_metrics_column_family_id
        )

        
        try:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Shared Metrics Column for Column Family - {self._shared_metrics_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._metrics_table_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Shared Metrics Column for Column Family - {self._shared_metrics_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Shared Metrics Column for Column Family - {self._shared_metrics_column_family_id} - if not exists')


        rows = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            row_key = f'{self.shared_metrics_table_id}_{str(uuid.uuid4())}'
            row = stage_metrics_table.direct_row(row_key)

            stage_metrics_record = {
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                'group': 'common',
                **metrics_set.common_stats
            }

            for field, value in stage_metrics_record.items():
                if not isinstance(value, bytes):
                    value = f'{value}'.encode()

                row.set_cell(
                    self._metrics_column_family_id,
                    field,
                    value,
                    timestamp=datetime.now()
                )

            rows.append(row)

        await self._loop.run_in_executor(
            self._executor,
            stage_metrics_table.mutate_rows,
            rows
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Table {self.shared_metrics_table_id}')
    
    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Table {self.metrics_table_id}')

        metrics_table = self.instance.table(self.metrics_table_id)
            
        try:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Metrics Column Family - {self._metrics_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                metrics_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Metrics Column Family - {self._metrics_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Metrics Column Family - {self._metrics_column_family_id} - if not exists')

        self._metrics_table_columns = metrics_table.column_family(
            self._metrics_column_family_id
        )
        
        try:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Metrics Column for Column Family - {self._metrics_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._metrics_table_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Metrics Column for Column Family - {self._metrics_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Metrics Column for Column Family - {self._metrics_column_family_id} - if not exists')
        
        rows = []
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Group - {group_name}:{group.metrics_group_id}')

                row_key = f'{self.metrics_table_id}_{str(uuid.uuid4())}'
                row = metrics_table.direct_row(row_key)

                record = {
                    'group': group_name,
                    **group.record
                }

                for field, value in record.items():
                    if not isinstance(value, bytes):
                        value = f'{value}'.encode()

                    row.set_cell(
                        self._metrics_column_family_id,
                        field,
                        value,
                        timestamp=datetime.now()
                    )

                rows.append(row)

        await self._loop.run_in_executor(
            self._executor,
            metrics_table.mutate_rows,
            rows
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Table {self.metrics_table_id}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to Table {self.metrics_table_id}')

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Group - Custom')

        custom_metrics_table = self.instance.table(self.custom_metrics_table_id)

        try:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Custom Metrics Column Family - {self._custom_metrics_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                custom_metrics_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Custom Metrics Column Family - {self._custom_metrics_column_family_id} - if not exists')
        
        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Custom Metrics Column Family - {self._custom_metrics_column_family_id} - if not exists')

        custom_metrics_table_columns = custom_metrics_table.column_family(
            self.custom_metrics_table_id
        )

        try:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Custom Metrics Column for Column Family - {self._custom_metrics_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                custom_metrics_table_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Custom Metrics Column for Column Family - {self._custom_metrics_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Custom Metrics Column for Column Family - {self._custom_metrics_column_family_id} - if not exists')

        rows = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            row_key = f'custom_{str(uuid.uuid4())}'
            custom_metrics_table_row = custom_metrics_table.direct_row(row_key)

            custom_metrics_row_data = {
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                'group': 'custom',
                **{
                    metric.metric_shortname: metric.metric_value for metric in metrics_set.custom_metrics.values()
                }
            }

            for field, value in custom_metrics_row_data.items():
                custom_metrics_table_row.set_cell(
                    field,
                    f'{value}'.encode()
                )
                
                rows.append(custom_metrics_table_row)

        await self._loop.run_in_executor(
            self._executor,
            custom_metrics_table.mutate_rows,
            rows
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Table {self.metrics_table_id}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Table {self.errors_table_id}')

        errors_table = self.instance.table(self.errors_table_id)
            
        try:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Error Metrics Column Family - {self._errors_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                errors_table.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Error Metrics Column Family - {self._errors_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Error Metrics Column Family - {self._errors_column_family_id} - if not exists')

        self._errors_table = errors_table

        self._errors_table_columns = errors_table.column_family(
            self._errors_column_family_id
        )

        try:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Error Column for Column Family - {self._errors_column_family_id} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                self._metrics_table_columns.create
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created Error Column for Column Family - {self._errors_column_family_id} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping creation of Error Column for Column Family - {self._errors_column_family_id} - if not exists')
        
        rows = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            for error in metrics_set.errors:

                row_key = f'{self.errors_table_id}_{str(uuid.uuid4())}'
                errors_row = errors_table.direct_row(row_key)

                error_message = error.get('message')
                error_count = error.get('count')

                errors_row.set_cell(
                    self._metrics_column_family_id,
                    'error_type',
                    f'{error_message}'.encode()
                )

                errors_row.set_cell(
                    self._metrics_column_family_id,
                    'error_count',
                    f'{error_count}'.encode()
                )

                rows.append(errors_row)
                
        await self._loop.run_in_executor(
            self._executor,
            errors_table.mutate_rows,
            rows
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Table {self.errors_table_id}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self._loop.run_in_executor(
            self._executor,
            self.client.close
        )

        self._executor.shutdown(cancel_futures=True)


        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Session Closed - {self.session_uuid}')
        


from .bidirectional_gzip_decompressor import BidirectionalGZipDecompressor
from .bidirectional_zstandard_decompressor import BidirectionalZStandardDecompressor
from .gzip_decompressor import GZipDecompressor
from .zstandard_decompressor import ZStandardDecompressor


from __future__ import annotations
import json
from typing import Dict, Union, Any
from gzip import decompress as gzip_decompress
from zlib import decompress as zlib_decompress
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common.base_result import BaseResult
from .events.deferred_headers_event import DeferredHeaders
from .action import HTTP2Action


class HTTP2Result(BaseResult):

    __slots__ = (
        'action_id',
        'url',
        'ip_addr',
        'method',
        'path',
        'params',
        'query',
        'hostname',
        'body',
        'response_code',
        'deferred_headers',
        '_headers',
        '_compression',
        '_content_type',
        '_size',
        '_version',
        '_reason',
        '_status'
    )

    def __init__(self, action: HTTP2Action, error: Exception = None) -> None:
        super(
            HTTP2Result,
            self
        ).__init__(
            action.action_id,
            action.name,
            action.url.hostname,
            action.metadata.user,
            action.metadata.tags,
            RequestTypes.HTTP2,
            error
        )

        self.url = action.url.full
        self.ip_addr = action.url.ip_addr
        self.method = action.method
        self.path = action.url.path
        self.params = action.url.params
        self.query = action.url.query
        self.hostname = action.url.hostname
        self._headers: Dict[bytes, bytes] = {}
        self.body = bytearray()
        
        self.response_code: str = None
        self.deferred_headers: DeferredHeaders = None
        self._compression = None
        self._content_type = None
        self._size = None
        self._version = None
        self._reason = None
        self._status = None

    @property
    def headers(self) -> Dict[bytes, bytes]:
        if self._headers is None or  len(self._headers) == 0:
            self._headers = self._parse_headers()
        
        return self._headers

    @headers.setter
    def headers(self, value: Any):
        self._headers = value

    @property
    def content_type(self):
        if len(self.headers) == 0 and self.deferred_headers:
            self.headers = self._parse_headers()
            self._content_type = self.headers.get(b'content-type')
        
        return self._content_type

    @content_type.setter
    def content_type(self, value: str):
        self._content_type = value

    @property
    def compression(self):
        if len(self.headers) == 0 and self.deferred_headers:
            self.headers = self._parse_headers()
            self._compression = self.headers.get(b'content-encoding')

        return self._compression

    @compression.setter
    def compression(self, value: str):
        self._compression = value

    @property
    def version(self) -> Union[str, None]:
        if len(self.headers) == 0 and self.deferred_headers:
            self.headers = self._parse_headers()
            self._version = self.headers.get(b'version')

        return self._version

    @version.setter
    def version(self, value: str):
        self._version = value

    @property
    def status(self) -> Union[int, None]:
        if self._status is None and len(self.headers) == 0 and self.deferred_headers:
            self.headers = self._parse_headers()
            self._status = int(self.headers.get(b'status', 400))
        
        return self._status

    @status.setter
    def status(self, value: int):
        self._status = value

    @property
    def reason(self) -> Union[str, None]:
        if len(self.headers) == 0 and self.deferred_headers:
            self.headers = self._parse_headers()
            self._reason = self.headers.get(b'reason')

        return self._reason

    @reason.setter
    def reason(self, value: str):
        self._reason = value

    @property
    def size(self):
        
        if len(self.headers) == 0 and self.deferred_headers:
            self.headers = self._parse_headers()
            content_length = self.headers.get(b'content-length')
            if content_length:
                self._size = int(content_length)

            elif len(self.body) > 0:
                self._size = len(self.body)
            
            else:
                self._size = 0

        return self._size

    @size.setter
    def size(self, value: int):
        self._size = value
        
    @property
    def data(self) -> Union[str, dict, None]:

        if len(self.headers) == 0 and self.deferred_headers:
            self._headers = self._parse_headers()

        data = self.body
        try:
            if self.headers.get(b'content-encoding') == b"gzip":
                data = gzip_decompress(self.body)
            elif self.headers.get(b'content-encoding') == b"deflate":
                data = zlib_decompress(self.body)

            if self.headers.get(b'content-type') == b"application/json":
                data = json.loads(self.body)
            
            elif isinstance(self.body, (bytes, bytearray)):
                data = str(self.body.decode())

        except Exception:
            pass

        return data

    @data.setter
    def data(self, value):
        self.body = value

    def _parse_headers(self):
        try:
            status, decoded_headers = self.deferred_headers.parse()
            decoded_headers['status'] = status
            return decoded_headers

        except Exception:
            return {}



from typing import (
    Callable, 
    Awaitable, 
    Any, 
    Tuple, 
    Dict,
    Union
)
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.common.action_registry import actions_registry
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.hook import Hook
from hedra.data.connectors.aws_lambda.aws_lambda_connector_config import AWSLambdaConnectorConfig
from hedra.data.connectors.bigtable.bigtable_connector_config import BigTableConnectorConfig
from hedra.data.connectors.cassandra.cassandra_connector_config import CassandraConnectorConfig
from hedra.data.connectors.cosmosdb.cosmos_connector_config import CosmosDBConnectorConfig
from hedra.data.connectors.csv.csv_connector_config import CSVConnectorConfig
from hedra.data.connectors.google_cloud_storage.google_cloud_storage_connector_config import GoogleCloudStorageConnectorConfig
from hedra.data.connectors.har.har_connector_config import HARConnectorConfig
from hedra.data.connectors.json.json_connector_config import JSONConnectorConfig
from hedra.data.connectors.kafka.kafka_connector_config import KafkaConnectorConfig
from hedra.data.connectors.mongodb.mongodb_connector_config import MongoDBConnectorConfig
from hedra.data.connectors.mysql.mysql_connector_config import MySQLConnectorConfig
from hedra.data.connectors.postgres.postgres_connector_config import PostgresConnectorConfig
from hedra.data.connectors.redis.redis_connector_config import RedisConnectorConfig
from hedra.data.connectors.s3.s3_connector_config import S3ConnectorConfig
from hedra.data.connectors.snowflake.snowflake_connector_config import SnowflakeConnectorConfig
from hedra.data.connectors.sqlite.sqlite_connector_config import SQLiteConnectorConfig
from hedra.data.connectors.xml.xml_connector_config import XMLConnectorConfig
from hedra.data.connectors.connector import Connector


class SaveHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        *names: Tuple[str, ...],
        loader: Union[
            AWSLambdaConnectorConfig,
            BigTableConnectorConfig,
            CassandraConnectorConfig,
            CosmosDBConnectorConfig,
            CSVConnectorConfig,
            GoogleCloudStorageConnectorConfig,
            HARConnectorConfig,
            JSONConnectorConfig,
            KafkaConnectorConfig,
            MongoDBConnectorConfig,
            MySQLConnectorConfig,
            PostgresConnectorConfig,
            RedisConnectorConfig,
            S3ConnectorConfig,
            SnowflakeConnectorConfig,
            SQLiteConnectorConfig,
            XMLConnectorConfig

        ]=None, 
        order: int=1,
        skip: bool=False
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            order=order,
            skip=skip,
            hook_type=HookType.SAVE
        )

        self.names = list(set(names))
        self.loader_config = loader
        self.parser_config: Union[Config, None] = None
        self.connector: Union[Connector, None] = Connector(
            self.stage,
            self.loader_config,
            self.parser_config
        )

        self.saved = False

    async def call(self, **kwargs) -> None:

        condition_result = await self._execute_call(**kwargs)

        if self.skip or self.saved or condition_result is False:
            return kwargs

        if self.connector.connected is False:
            self.connector.selected.stage = self.stage
            self.connector.selected.parser_config = self.parser_config

            await self.connector.connect()

        hook_args = {
            name: value for name, value in kwargs.items() if name in self.params
        }
        
        load_result: Union[Dict[str, Any], Any] = await self._call(**{
            **hook_args,
            'actions': actions_registry.actions(),
            'connector': self.connector
        })
        
        await self.connector.close()


        self.saved = True

        if isinstance(load_result, dict):
            return {
                **kwargs,
                **load_result
            }

        return {
            **kwargs,
            self.shortname: load_result
        }

    def copy(self):
        save_hook = SaveHook(
            self.name,
            self.shortname,
            self._call,
            *self.names,
            loader=self.loader_config,
            order=self.order,
            skip=self.skip,
        )

        save_hook.stage = self.stage

        return save_hook


from typing import Generic
from typing_extensions import TypeVarTuple, Unpack
from hedra.core.engines.client import Client
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.internal.decorator import Internal
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.base.parallel.stage_priority import StagePriority
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.monitoring import (
    CPUMonitor,
    MemoryMonitor
)
from typing import Optional


T = TypeVarTuple('T')


class Act(Stage, Generic[Unpack[T]]):
    stage_type=StageTypes.ACT
    priority: Optional[str]=None
    retries=0

    def __init__(self) -> None:
        super().__init__()
        self.persona = None
        self.client: Client[Unpack[T]] = Client(
            self.graph_name,
            self.graph_id,
            self.name,
            self.stage_id
        )
        
        self.accepted_hook_types = [ 
            HookType.ACTION,
            HookType.CHANNEL, 
            HookType.CHECK,
            HookType.CONDITION,
            HookType.CONTEXT,
            HookType.EVENT,
            HookType.LOAD,
            HookType.SAVE,
            HookType.TASK,
            HookType.TRANSFORM
        ]

        self.priority = self.priority
        if self.priority is None:
            self.priority = 'auto'

        self.priority_level: StagePriority = StagePriority.map(
            self.priority
        )

        self.stage_retries = self.retries

    @Internal()
    async def run(self):
        await self.setup_events()
        self.dispatcher.assemble_execution_graph()

        cpu_monitor = CPUMonitor()
        memory_monitor = MemoryMonitor()

        main_monitor_name = f'{self.name}.main'

        await cpu_monitor.start_background_monitor(main_monitor_name)
        await memory_monitor.start_background_monitor(main_monitor_name)

        await self.dispatcher.dispatch_events(self.name)

        await cpu_monitor.stop_background_monitor(main_monitor_name)
        await memory_monitor.stop_background_monitor(main_monitor_name)

        cpu_monitor.close()
        memory_monitor.close()

        cpu_monitor.stage_metrics[main_monitor_name] = cpu_monitor.collected[main_monitor_name]
        memory_monitor.stage_metrics[main_monitor_name] = memory_monitor.collected[main_monitor_name]

        self.context.update({
            'act_stage_monitors': {
                self.name: {
                    'cpu': cpu_monitor,
                    'memory': memory_monitor
                }
            }
        })



